// naumann@stce.rwth-aachen.de

#include "dco.hpp"

#include <iostream>
#include <fstream>
#include <cmath>
using namespace std;

double eps=1e-10;

template<typename T>
T f(T x) {
  return pow(x,2);
  return exp(sin(pow(x,2)));
}

template<typename T>
T df_dx(T x) {
  using TT=dco::gt1s<T>::type;
  TT x_t, y_t;
  dco::value(x_t)=x;
  dco::derivative(x_t)=1;
  y_t=f(x_t);
  return dco::derivative(y_t);
  // T delta_x=1e-7;
  // return (f(x+delta_x)-f(x-delta_x))/(2*delta_x);
  // return exp(sin(pow(x,2)))*cos(pow(x,2))*2*x;
}

template<typename T>
T ddf_dxx(T x) {
  using TT=dco::gt1s<T>::type;
  TT x_t, y_t;
  dco::value(x_t)=x;
  dco::derivative(x_t)=1;
  y_t=df_dx(x_t);
  return dco::derivative(y_t);
  // T delta_x=1e-7;
  // return (df_dx(x+delta_x)-df_dx(x-delta_x))/(2*delta_x);
  // return exp(sin(pow(x,2)))*cos(pow(x,2))*2*x*cos(pow(x,2))*2*x+
  // exp(sin(pow(x,2)))*(-sin(pow(x,2)))*2*x*2*x+
  // exp(sin(pow(x,2)))*cos(pow(x,2))*2;
}

template<typename T>
T dddf_dxxx(T x) {
  using TT=dco::gt1s<T>::type;
  TT x_t, y_t;
  dco::value(x_t)=x;
  dco::derivative(x_t)=1;
  y_t=ddf_dxx(x_t);
  return dco::derivative(y_t);
}

template<typename T>
T preprocess(T x) {
  T dfdx=df_dx(x);
  do {
    cout << "-";
    T alpha=1;
    T y=f(x);
    while (f(x-alpha*dfdx)>=y) {
      alpha=alpha/2;
    }
    x=x-alpha*dfdx;
    dfdx=df_dx(x);
  } while (fabs(dfdx)>0.1);
  return x;
}

template<typename T>
T minimize(T x) {
  T dfdx=df_dx(x);
  do {
    cout << "+";
    x=x-dfdx/ddf_dxx(x);
    dfdx=df_dx(x);
  } while (fabs(dfdx)>eps);
  return x;
}

int main() {
  double x=1.5;
  cout << "x=" << x << endl;
  cout << "f(" << x << ")=" << f(x) << endl;
  x=preprocess(x);
  x=minimize(x);
  cout << endl;
  cout << "x=" << x << endl;
  cout << "f(" << x << ")=" << f(x) << endl;
  cout << "dfdx(" << x << ")=" << df_dx(x) << endl;
  cout << "ddfdxx(" << x << ")=" << ddf_dxx(x) << endl;
  return 0;
}


