// naumann@stce.rwth-aachen.de

#include <Eigen/Dense>
template<typename T>
using VT=Eigen::VectorX<T>;

#include "dco.hpp"

#include <cmath>
#include <cassert>
#include <fstream>
#include <iostream>
using namespace std;

template<typename T>
T f(T p, double x) {
  assert(x>=0);
  return p*sin(x);
}

double e(double p, VT<double> xobs, VT<double> yobs) {
  assert(xobs.size()==yobs.size());
  double e=0;
  int i=0;
  while (i<xobs.size()) {
    e=e+pow(f(p,xobs(i))-yobs(i),2);
    i=i+1;
  }
  return e;
}

template<typename T>
VT<T> r(T p, VT<double> xobs, VT<double> yobs) {
  assert(xobs.size()==yobs.size());
  VT<T> r(xobs.size());
  int i=0;
  while (i<xobs.size()) {
    r(i)=f(p,xobs(i))-yobs(i);
    i=i+1;
  }
  return r;
}

VT<double> dr_dp(double p, VT<double> xobs, VT<double> yobs) {
  using TT=dco::gt1s<double>::type;
  VT<TT> r_t(xobs.size());
  TT p_t;
  dco::value(p_t)=p;
  dco::derivative(p_t)=1;
  r_t=r(p_t,xobs,yobs);
  VT<double> drdp(xobs.size());
  int i=0;
  while (i<xobs.size()) {
    drdp(i)=dco::derivative(r_t(i));
    i=i+1;
  }
  return drdp;
}

double fit(VT<double> xobs, VT<double> yobs) {
  assert(xobs.size()==yobs.size());
  double p;
  VT<double> rp=dr_dp(p,xobs,yobs);
  return rp.dot(yobs)/rp.dot(rp);
}

int main() {
  double p=0.5;
  int nobs=10;
  VT<double> xobs=VT<double>::Random(nobs); 
  xobs=xobs.cwiseProduct(xobs); 
  VT<double> yobs=VT<double>::Random(nobs); 
  ofstream ofs("data");
  int i=0;
  while (i<nobs) {
    ofs << xobs(i) << ' ' << yobs(i) << endl;
    i=i+1;
  }
  cout << "p=" << p << endl;
  cout << "e=" << e(p,xobs,yobs) << endl;
  cout << "dedp=" << (e(p+1e-7,xobs,yobs)-e(p-1e-7,xobs,yobs))/2e-7 << endl;
  p=fit(xobs,yobs);
  cout << "p=" << p << endl;
  cout << "e=" << e(p,xobs,yobs) << endl;
  cout << "dedp=" << (e(p+1e-7,xobs,yobs)-e(p-1e-7,xobs,yobs))/2e-7 << endl;
  return 0;
}
